Uses Crt,Graph;

Const MaxPop = 35;
      ChromSize = 7;
      DeadColor = 2;

Type
     PGenotype = ^TGenotype;
     TGenotype = record
       case Integer of
       0: (MaxSpeed,BounceSpeed,Acceleration : Real;
           DistBrotherFactor,DistStrangerFactor : Real;
           CollisionBrotherFactor,CollisionStrangerFactor : Real);
       1: (C:Array[0..ChromSize-1] of Real);
     End;

     PEnvironmentParam = ^TEnvironmentParam;
     TEnvironmentParam = record
       width,height : LongInt;
       EnergyFactor,SafetyFactor,SurviversFactor : Real;
       MutationFactor : Real;
       MaxEnergyDose,MaxSafetyDose : Real;
       CenterAcceleration : Real;
       PopSize : LongInt;
       NumOfNeighbors : Longint;
       CollisionDistance : LongInt;
       DistLocalFactor,CollisionLocalFactor : Real;
     End;

     PFixedParam = ^TFixedParam;
     TFixedParam = record
       energy,safety : LongInt;
       Color,typ : Byte;
     End;

     PFloy = ^TFloy;

     PFloyArray = ^TFloyArray;
     TFloyArray = Array[1..MaxPop] of PFloy;

     TFloy = object
       Neighbors : PFloyArray;
       DadID,MomID,typ : LongInt;
       Color : Byte;
       Fitness,Energy,Safety : LongInt;
       x,y,xtail,ytail,vx,vy : Real;
       env : PEnvironmentParam;
       chrom : PGenotype;
       Constructor InitRandom(fix : PFixedParam;ev : PEnvironmentParam);
       Constructor InitFixed(cr : TGenotype;fix : PFixedParam;ev : PEnvironmentParam);
       Constructor InitCrossOver(mom,dad : PFloy;fix : PFixedParam;ev : PEnvironmentParam);
       Procedure GetFitness;
       Procedure Mutate;
       Function Dist(Floy : PFloy) : LongInt;
       Procedure Draw;
       Procedure GetNeighbors;
       Procedure Process;
       Destructor Done;
     End;

     TEvolution = object
       Floys,NewFloys : PFloyArray;
       Env : TEnvironmentParam;
       Fix1,Fix2 : TFixedParam;
       minc,maxc : TGenotype; { for mutation }
       Chrom : TGenotype;
       BeforeEvol : Pointer;
       ExitSimulation : Boolean;
       NewGeneration : Boolean;
       ExitProgram : Boolean;
       AutoMode : Boolean;
       ShowLeaders : Boolean;
       SumFitness,AvgFitness : Real;
       SumSpeed,AvgSpeed : Real;
       SumAcc,AvgAcc : Real;
       SumDll,AvgDll : Real;
       SumCll,AvgCll : Real;
       procedure InitParams;
       procedure Rand;
       function GetSumFitness : Real;
       function GetParentID : LongInt;
       procedure Mate(Var Kid : PFloy);
       procedure SortFloys(Var F : PFloyArray);
       procedure RecordGeneration;
       procedure ReleaseStranger;
       procedure Breed;
       procedure InitAll;
       procedure Move;
       procedure ShowLeadersStats;
       procedure ShowMenu;
     End;

     TEvolApplication = object
       Gd,Gm,ErrCode : Integer;
       CurPage : LongInt;
       Time : LongInt;
       constructor Init;
       procedure Run;
       destructor Done;
     End;

     TNoMouse = object
       constructor Reset;
       destructor Done;
       procedure Show; virtual;
       procedure Hide; virtual;
       procedure GetState; virtual;
       procedure SetPos(xx,yy : integer); virtual;
       procedure SetXLimit(minx,maxx : integer); virtual;
       procedure SetYLimit(miny,maxy : integer); virtual;
     End;

     TMouse = object (TNoMouse)
       x,y : integer;
       rb,lb : boolean;
       constructor Reset;
       procedure Show; virtual;
       procedure Hide; virtual;
       procedure GetState; virtual;
       procedure SetPos(xx,yy : integer); virtual;
       procedure SetXLimit(minx,maxx : integer); virtual;
       procedure SetYLimit(miny,maxy : integer); virtual;
     end;

Var
        Timer : LongInt absolute $40:$6c;
        Mouse : TMouse;
        Evol  : TEvolution;
        App   : TEvolApplication;

Procedure WaitVerticalRetrace; assembler;
asm
  mov dx,3DAh
@VRT1:
  in al,dx
  test al,8
  jnz @VRT1
@VRT2:
  in al,dx
  test al,8
  jz @VRT2
end;

{$L ev}
Procedure EgaVga; external; { egavga.bgi file }

Function Flip(Probability : Real) : Boolean;
Begin
  Flip:=Random<Probability;
End;

Function IntToStr(L : LongInt) : String;
Var S : String;
Begin
  Str(L,S);
  IntToStr:=S;
End;

Function RealToStr(R : Real) : String;
Var S : String;
Begin
  Str(R:0:2,S);
  RealToStr:=S;
End;

Constructor TNoMouse.Reset;
Begin
  Reset; { this instruction will never be executed, but it must be here ;)
           It is highly UNRECOMMENDED to remove this recursive call - make
           it on your own risk }
End;

Destructor TNoMouse.Done;
Begin
End;

procedure TNoMouse.Show;
Begin
End;

procedure TNoMouse.Hide;
Begin
End;

procedure TNoMouse.GetState;
Begin
End;

procedure TNoMouse.SetPos;
Begin
End;

procedure TNoMouse.SetXLimit;
Begin
End;

procedure TNoMouse.SetYLimit;
Begin
End;

Constructor TMouse.Reset; assembler;
asm
  xor ax,ax
  int 33h             { ax=$FFFF if mouse present }
  inc ax
  jz @skip            { mouse present }
  lds di,Self
  mov ax,$20
  sub [di],ax         { if mouse not present, set object type to TNoMouse }
  jmp @skip
  call TNoMouse.Reset { this instruction will never be executed,
                        but it must be here ;) It is highly UNRECOMMENDED to
                        remove this call - make it on your own risk }
@skip:
end;

procedure TMouse.Show; assembler;
asm
  xor ax,ax
  inc ax
  int 33h
end;

procedure TMouse.Hide; assembler;
asm
  mov ax,2
  int 33h
end;

procedure TMouse.GetState; assembler;
asm
  mov ax,3
  int 33h
  lds di,self
  mov [di+X],cx
  mov [di+Y],dx
  mov al,bl
  shr al,1
  mov [di+lb],al
  and bl,1
  mov [di+rb],bl
end;

procedure TMouse.SetPos; assembler;
asm
  mov ax,4
  mov cx,xx
  mov dx,yy
  int 33h
end;

procedure TMouse.SetXLimit; assembler;
asm
  mov ax,7
  mov cx,minx
  mov dx,maxx
  int 33h
end;

procedure TMouse.SetYLimit; assembler;
asm
  mov ax,8
  mov cx,miny
  mov dx,maxy
  int 33h
end;

constructor TFloy.InitRandom;
Var gene : PGenotype absolute chrom;
    i : LongInt;
Begin
  typ:=fix^.typ;
  color:=fix^.color;
  energy:=fix^.energy;
  safety:=fix^.safety;
  New(chrom);
  For i:=0 to ChromSize-1 do
    gene^.c[i]:=evol.minc.c[i]+Random*(evol.maxc.c[i]-evol.minc.c[i]);
  env:=ev;
  x:=Random(env^.width);
  y:=Random(env^.height);
  xtail:=x;
  ytail:=y;
  vx:=Random*chrom^.MaxSpeed-chrom^.MaxSpeed/2;
  vy:=Random*chrom^.MaxSpeed-chrom^.MaxSpeed/2;
  GetMem(Neighbors,4*env^.NumOfNeighbors);
End;

constructor TFloy.InitFixed;
Begin
  typ:=fix^.typ;
  color:=fix^.color;
  energy:=fix^.energy;
  safety:=fix^.safety;
  New(chrom);
  chrom^:=cr;
  env:=ev;
  x:=Random(env^.width);
  y:=Random(env^.height);
  xtail:=x;
  ytail:=y;
  vx:=Random*chrom^.MaxSpeed-chrom^.MaxSpeed/2;
  vy:=Random*chrom^.MaxSpeed-chrom^.MaxSpeed/2;
  GetMem(Neighbors,4*env^.NumOfNeighbors);
End;

constructor TFloy.InitCrossOver;
Var i,j : LongInt;
    gene : PGenotype absolute chrom;
Begin
  typ:=fix^.typ;
  color:=fix^.color;
  energy:=fix^.energy;
  safety:=fix^.safety;
  New(chrom);
  i:=Longint(Random(ChromSize))-1;
  For j:=0 to i do
    gene^.c[j]:=mom^.chrom^.c[j];
  For j:=i+1 to ChromSize-1 do
    gene^.c[j]:=dad^.chrom^.c[j];
  env:=ev;
  x:=Random(env^.width);
  y:=Random(env^.height);
  xtail:=x;
  ytail:=y;
  vx:=Random*chrom^.MaxSpeed-chrom^.MaxSpeed/2;
  vy:=Random*chrom^.MaxSpeed-chrom^.MaxSpeed/2;
  GetMem(Neighbors,4*env^.NumOfNeighbors);
End;

Procedure TFloy.GetFitness;
Begin
  fitness:=Trunc(energy*env^.EnergyFactor+safety*env^.SafetyFactor);
End;

Procedure TFloy.Mutate;
Var i : LongInt;
    Gene : PGenotype absolute Chrom;
Begin
  For i:=0 to ChromSize-1 do
  Begin
    If Flip(env^.MutationFactor) then
      Gene^.c[i]:=Random*(evol.maxc.c[i]-evol.minc.c[i])+evol.minc.c[i];
  End;
End;

Function TFloy.Dist;
var d1,d2 : LongInt;
Begin
  d1:=Trunc(x-Floy^.x);
  d2:=Trunc(y-Floy^.y);
  Dist:=d1*d1+d2*d2;
End;

Procedure TFloy.Draw;
Begin
  If x<2 then x:=2;
  If x>env^.width-3 then x:=env^.width-3;
  If y<2 then y:=2;
  If y>env^.height-3 then y:=env^.height-3;
  SetColor(Color);
  Line(Trunc(xtail),Trunc(ytail),Trunc(x),Trunc(y));
  Circle(Trunc(x),Trunc(y),2);
  Circle(Trunc(x),Trunc(y),1);
End;

Procedure TFloy.GetNeighbors;
Var i,j : LongInt;
    d : LongInt;
    F : PFloy;
Begin
  For i:=1 to env^.NumOfNeighbors do
  Begin
    F:=Neighbors^[i];
    d:=Dist(F);
    For j:=1 to env^.NumOfNeighbors do
    Begin
      If Dist(F^.Neighbors^[j])<d then
        Neighbors^[i]:=F^.Neighbors^[j];
      If F^.Neighbors^[j]^.typ=1 then
        Neighbors^[i]:=F^.Neighbors^[j];
    End;
  End;
  For i:=1 to env^.NumOfNeighbors do
  Begin
    If Neighbors^[i]=@Self then Neighbors^[i]:=Neighbors^[i]^.Neighbors^[1];
  End;
End;

Procedure TFloy.Process;
Var i,j : LongInt;
    d : LongInt;
    rev : LongInt;
Begin
  xtail:=x;
  ytail:=y;
  rev:=-1;

  For i:=1 to env^.NumOfNeighbors do
  Begin
    d:=Dist(Neighbors^[i]);
    If d=0 then rev:=0 else
      If d<env^.CollisionDistance then
        If Neighbors^[i]^.typ<>typ then
          If typ=1 then
          Begin
            Dec(energy);
            rev:=Trunc(env^.CollisionLocalFactor);
            If energy<0 then
            Begin
              energy:=0;
              Color:=DeadColor;
              evol.NewGeneration:=True;
            End;
          End
          else
          Begin
            Inc(energy,Trunc(env^.MaxEnergyDose));
            rev:=Trunc(chrom^.CollisionStrangerFactor);
          End
        else
        Begin
          Inc(Safety,Trunc(env^.MaxSafetyDose));
          rev:=Trunc(chrom^.CollisionBrotherFactor);
        End
      else
      Begin
        If typ=1 then rev:=Trunc(env^.DistLocalFactor) else
          If Neighbors^[i]^.typ=1 then rev:=Trunc(chrom^.DistStrangerFactor) else
          Begin
            Inc(Safety,Trunc((env^.MaxSafetyDose*env^.CollisionDistance)/(d+env^.CollisionDistance)));
            rev:=Trunc(chrom^.DistBrotherFactor);
          End;
      End;

    If x<Neighbors^[i]^.x then vx:=vx+chrom^.Acceleration*rev else
      vx:=vx-chrom^.Acceleration*rev;
    If y<Neighbors^[i]^.y then vy:=vy+chrom^.Acceleration*rev else
      vy:=vy-chrom^.Acceleration*rev;
  End;

  If vx>chrom^.MaxSpeed then vx:=chrom^.MaxSpeed;
  If vx<-chrom^.MaxSpeed then vx:=-chrom^.MaxSpeed;
  If vy>chrom^.MaxSpeed then vy:=chrom^.MaxSpeed;
  If vy<-chrom^.MaxSpeed then vy:=-chrom^.MaxSpeed;

  If x<10 then vx:=chrom^.BounceSpeed;
  If x>env^.width-10 then vx:=-chrom^.BounceSpeed;
  If y<10 then vy:=chrom^.BounceSpeed;
  If y>env^.height-10 then vy:=-chrom^.BounceSpeed;

  If typ=0 then
  Begin
    Dec(energy,Trunc(Sqrt(vx*vx+vy*vy)*env^.MaxEnergyDose));
    If x<env^.width/2 then vx:=vx+env^.CenterAcceleration;
    If x>env^.width/2 then vx:=vx-env^.CenterAcceleration;
    If y<env^.height/2 then vy:=vy+env^.CenterAcceleration;
    If y>env^.height/2 then vy:=vy-env^.CenterAcceleration;
  End;

  x:=x+vx;
  y:=y+vy;
End;

Destructor TFloy.Done;
Begin
  FreeMem(Neighbors,4*env^.NumOfNeighbors);
  Dispose(chrom);
End;

Procedure TEvolution.InitParams;
Begin
  With Env do
  Begin
    PopSize:=20;
    width:=480;
    height:=350;
    MaxEnergyDose:=10;
    MaxSafetyDose:=10;
    EnergyFactor:=-5;
    SafetyFactor:=0;
    SurviversFactor:=0.1;
    CollisionDistance:=200;
    CenterAcceleration:=0.01;
    NumOfNeighbors:=2;
    DistLocalFactor:=0;
    CollisionLocalFactor:=-40;
    MutationFactor:=0.3;
  End;
  With Fix1 do
  Begin
    Color:=10;
    typ:=0;
    energy:=10;
    safety:=10;
  End;
  With Fix2 do
  Begin
    Color:=13;
    typ:=1;
    energy:=10;
    safety:=10;
  End;
  With maxc do
  Begin
    MaxSpeed:=8;
    BounceSpeed:=2;
    Acceleration:=0.8;
    DistBrotherFactor:=3;
    DistStrangerFactor:=40;
    CollisionBrotherFactor:=-1;
    CollisionStrangerFactor:=50;
  End;
  With minc do
  Begin
    MaxSpeed:=1;
    BounceSpeed:=0.1;
    Acceleration:=0.05;
    DistBrotherFactor:=0;
    DistStrangerFactor:=10;
    CollisionBrotherFactor:=-5;
    CollisionStrangerFactor:=10;
  End;
  With Chrom do
  Begin
    MaxSpeed:=5;
    BounceSpeed:=0.8;
    Acceleration:=0.3;
    DistBrotherFactor:=1;
    DistStrangerFactor:=20;
    CollisionBrotherFactor:=-1;
    CollisionStrangerFactor:=30;
  End;
  AutoMode:=False;
  ShowLeaders:=True;
  ExitProgram:=False;
End;

Procedure TEvolution.Rand;
Var i,j : LongInt;
Begin
  For i:=1 to env.PopSize do
    For j:=1 to env.NumOfNeighbors do
    Begin
      Floys^[i]^.Neighbors^[j]:=Floys^[Random(env.PopSize)+1];
    End;
End;

Function TEvolution.GetSumFitness;
Var i,min : LongInt;
    sum,avg : Real;
Begin
{ make all energy positive }
  min:=MaxLongInt;
  sum:=0;
  For i:=1 to env.PopSize do
  Begin
    Floys^[i]^.GetFitness;
    If Floys^[i]^.fitness<min then min:=Floys^[i]^.fitness;
    sum:=Floys^[i]^.fitness;
  End;
  If min<0 then
  Begin
    min:=-min; { make it positive }
    For i:=1 to env.PopSize do
      Inc(Floys^[i]^.fitness,min);
    sum:=min*env.PopSize;
  End;
  AvgFitness:=sum/env.PopSize;
  GetSumFitness:=sum;
End;

Function TEvolution.GetParentID;
Var i,j,k,rnd,partsum : LongInt;
Begin
  partsum:=0;
  rnd:=Trunc(Random*SumFitness);
  j:=Env.PopSize;
  For i:=1 to env.PopSize do
  Begin
    Inc(partsum,Floys^[i]^.fitness);
    If partsum>=rnd then
    Begin
      j:=i;
      Break;
    End;
  End;
  GetParentID:=j;
End;

Procedure TEvolution.Mate;
Var MomID,DadID : LongInt;
    Mom,Dad : PFloy;
Begin
  Repeat
    MomID:=GetParentID;
  Until Floys^[MomID]^.typ<>1;
  Mom:=Floys^[MomID];
  Repeat
    DadID:=GetParentID;
  Until Floys^[DadID]^.typ<>1;
  Dad:=Floys^[DadID];
  New(Kid,InitCrossOver(Mom,Dad,@fix1,@env));
  Kid^.Mutate;
  Kid^.MomID:=MomID;
  Kid^.DadID:=DadID;
End;

Procedure TEvolution.SortFloys;
Var i,j : LongInt;
    TF : PFloy;
Begin
  For i:=1 to env.PopSize-1 do
    For j:=i+1 to env.PopSize do
      If F^[i]^.fitness<F^[j]^.fitness then
      Begin
        TF:=F^[i];
        F^[i]:=F^[j];
        F^[j]:=TF;
      End;
End;

Procedure TEvolution.RecordGeneration;
Var i : LongInt;
Begin
  SumSpeed:=0;
  SumAcc:=0;
  SumDll:=0;
  SumCll:=0;
  For i:=1 to env.PopSize do
  Begin
    SumSpeed:=SumSpeed+Floys^[i]^.chrom^.MaxSpeed;
    SumAcc:=SumAcc+Floys^[i]^.chrom^.Acceleration;
    SumDll:=SumDll+Floys^[i]^.chrom^.DistBrotherFactor;
    SumCll:=SumCll+Floys^[i]^.chrom^.CollisionBrotherFactor;
  End;
  AvgSpeed:=SumSpeed/env.PopSize;
  AvgAcc:=SumAcc/env.PopSize;
  AvgDll:=SumDll/env.PopSize;
  AvgCll:=SumCll/env.PopSize;
End;

Procedure TEvolution.ReleaseStranger;
Begin
  If env.PopSize<MaxPop then
  Begin
    Inc(Env.PopSize);
    New(Floys^[Env.PopSize],InitFixed(Chrom,@Fix2,@Env));
  End;
  Rand;
End;

Procedure TEvolution.Breed;
Var i,n : LongInt;
Begin
  While Floys^[env.PopSize]^.typ=1 do
  Begin
    Dispose(Floys^[env.PopSize],Done);
    Dec(env.PopSize);
  End;
  SumFitness:=GetSumFitness;
  For i:=1 to env.PopSize do
  Begin
    Mate(NewFloys^[i]);
  End;
  If env.SurviversFactor>0 then
  Begin
    SortFloys(Floys);
    n:=Trunc(env.PopSize*env.SurviversFactor);
    For i:=1 to n do
    Begin
      Dispose(NewFloys^[i],Done);
      Floys^[i]^.energy:=fix1.energy;
      Floys^[i]^.safety:=fix1.safety;
    End;
    For i:=n+1 to env.PopSize do
    Begin
      Dispose(Floys^[i],Done);
      Floys^[i]:=NewFloys^[i];
    End;
  End else
    For i:=1 to env.PopSize do
    Begin
      Dispose(Floys^[i],Done);
      Floys^[i]:=NewFloys^[i];
    End;
  Rand;
  RecordGeneration;
  NewGeneration:=False;
  If AutoMode then ReleaseStranger;
End;

Procedure TEvolution.InitAll;
Var i : LongInt;
Begin
  New(Floys);
  New(NewFloys);
  For i:=1 to Env.PopSize do
  Begin
    New(Floys^[i],InitRandom(@Fix1,@Env));
    Floys^[i]^.Mutate;
    Floys^[i]^.MomID:=0;
    Floys^[i]^.DadID:=0;
  End;
  Rand;
  RecordGeneration;
  NewGeneration:=False;
  If AutoMode then ReleaseStranger;
End;

Procedure TEvolution.Move;
Var i,j : LongInt;
Begin
  For i:=1 to Env.PopSize do
  Begin
    Floys^[i]^.GetNeighbors;
    Floys^[i]^.Process;
    Floys^[i]^.Draw;
  End;
  If Flip(0.05) then Rand;
  If KeyPressed then
  Case UpCase(ReadKey) of
  ' ': ReleaseStranger;
  'A': AutoMode:=not AutoMode;
  'B': Breed;
  #27: ExitSimulation:=True;
  End;
  j:=0;
  i:=1;
  While i<=Env.PopSize do
  If Floys^[i]^.color=DeadColor then
  Begin
    j:=1;
    Dispose(Floys^[i],Done);
    Floys^[i]:=Floys^[Env.PopSize];
    Dec(Env.PopSize);
  End else Inc(i);
  If j=1 then Rand;
  If NewGeneration and AutoMode then Breed;
End;

Procedure TEvolution.ShowLeadersStats;
Const Mode : Array[Boolean] of String = ('OFF','ON');
Var fit : Array[0..MaxPop] of record f,id:LongInt end;
    i,j,n,min : LongInt;
    F : PFloy;
Begin
  min:=MaxLongInt;
  n:=env.PopSize;
  For i:=1 to n do
  Begin
    Floys^[i]^.GetFitness;
    fit[i].f:=Floys^[i]^.fitness;
    fit[i].id:=i;
    If fit[i].f<min then min:=fit[i].f;
  End;
  If min<0 then
    For i:=1 to n do
      Dec(fit[i].f,min); { make all positive }
{ Sort fitness }
  For i:=1 to n-1 do
    For j:=i+1 to n do
      If fit[i].f<fit[j].f then
      Begin
        fit[0]:=fit[i];
        fit[i]:=fit[j];
        fit[j]:=fit[0];
      End;
  SetColor(Yellow);
  OutTextXY(480,0,'Spd  Acc  Dll  Cll');
  For i:=1 to n do
  Begin
    F:=Floys^[fit[i].id];
    OutTextXY(480,i*8,RealToStr(F^.chrom^.MaxSpeed)+' '+
    RealToStr(F^.chrom^.Acceleration)+' '+
    RealToStr(F^.chrom^.DistBrotherFactor)+' '+
    RealToStr(F^.chrom^.CollisionBrotherFactor));
  End;
  OutTextXY(480,(n+2)*8,'Averages:');
  OutTextXY(480,(n+3)*8,RealToStr(AvgSpeed)+' '+RealToStr(AvgAcc)+
  ' '+RealToStr(AvgDll)+' '+RealToStr(AvgCll));

  SetColor(White);
  OutTextXY(480,326,'SPACE = Add Stranger');
  OutTextXY(480,334,'A = AutoMode ('+Mode[AutoMode]+')');
  OutTextXY(480,342,'B = Next Generation');
End;

Procedure TEvolution.ShowMenu;
Const MenuHotKeys : Array[1..8] of Char = ('P','N','E','A','U','R','F','M');
Var i,j : LongInt;
    MenuItems : Array[1..8] of Pointer;
    MenuMin,MenuMax : Array[1..8] of Real;
    MenuTypes : Array[1..8] of (_INT,_FLOAT);
    MenuDigits : Array[1..8] of Byte;
    S : String;
    NextKey : Char;
    select : LongInt;
    Temp : Real;
    ErrCode : Integer;
Begin
  MenuTypes[1]:=_INT;
  MenuTypes[2]:=_INT;
  For i:=3 to 8 do MenuTypes[i]:=_FLOAT;
  MenuDigits[1]:=0;
  MenuDigits[2]:=0;
  MenuDigits[3]:=0;
  MenuDigits[4]:=0;
  For i:=5 to 8 do MenuDigits[i]:=3;
  MenuItems[1]:=@env.PopSize;
  MenuItems[2]:=@env.NumOfNeighbors;
  MenuItems[3]:=@env.EnergyFactor;
  MenuItems[4]:=@env.SafetyFactor;
  MenuItems[5]:=@env.SurviversFactor;
  MenuItems[6]:=@env.MaxEnergyDose;
  MenuItems[7]:=@env.MaxSafetyDose;
  MenuItems[8]:=@env.MutationFactor;
  MenuMin[1]:=2; MenuMax[1]:=MaxPop;
  MenuMin[2]:=1; MenuMax[2]:=10;
  MenuMin[3]:=-100000; MenuMax[3]:=100000;
  MenuMin[4]:=-100000; MenuMax[4]:=100000;
  MenuMin[5]:=0; MenuMax[5]:=1;
  MenuMin[6]:=-100000; MenuMax[6]:=100000;
  MenuMin[7]:=-100000; MenuMax[7]:=100000;
  MenuMin[8]:=0; MenuMax[8]:=1;

  SetVisualPage(0);
  SetActivePage(0);
  ClearDevice;
  SetColor(White);
  Line(476,0,476,349);
  OutTextXY(500,0,'Population Size:');
  OutTextXY(500,1,'_');
  OutTextXY(500,28,'Num of neighbors:');
  OutTextXY(500,29,'_');
  OutTextXY(500,56,'Energy factor:');
  OutTextXY(500,57,'_');
  OutTextXY(500,84,'Safety factor:');
  OutTextXY(508,85,'_');
  OutTextXY(500,112,'Survivers percent:');
  OutTextXY(508,113,'_');
  OutTextXY(500,140,'Max energy dose:');
  OutTextXY(556,141,'_');
  OutTextXY(500,168,'Max safety dose:');
  OutTextXY(548,169,'_');
  OutTextXY(500,196,'Mutation factor:');
  OutTextXY(500,197,'_');
  SetColor(LightGreen);
  OutTextXY(500,260,'START EVOLUTION');
  OutTextXY(500,261,'_');
  Rectangle(496,256,624,270);
  SetColor(LightRed);
  OutTextXY(490,290,'ESC = EXIT PROGRAM');
  Rectangle(486,286,638,300);

  SetColor(White);
  For i:=1 to 8 do
  Begin
    Case MenuTypes[i] of
    _INT: Str(LongInt(MenuItems[i]^),S);
    _FLOAT: Str(Real(MenuItems[i]^):0:MenuDigits[i],S);
    End;
    OutTextXY(540,i*28-14,S);
    Rectangle(536,i*28-18,544+Length(S)*8,i*28-4);
  End;

  Mouse.Show;

  Repeat
    Repeat
      Mouse.GetState;
    Until Mouse.lb or Mouse.rb or KeyPressed;
    select:=0;
    If Mouse.lb or Mouse.rb then
    Begin
      For i:=1 to 8 do
        If (Mouse.y>=i*28-28) and (Mouse.y<i*28) then
          select:=i;
      If (Mouse.y>=256) and (Mouse.y<=270) then select:=9; { START}
      If (Mouse.y>=286) and (Mouse.y<=300) then select:=10; { EXIT }
    End;
    If KeyPressed then
    Begin
      NextKey:=UpCase(ReadKey);
      Case NextKey of
      'S' : select:=9;
      #27 : select:=10;
      End;
      For i:=1 to 8 do
        If NextKey=MenuHotKeys[i] then select:=i;
    End;
    If (select>=1) and (select<=8) then
    Begin
      Mouse.Hide;
      i:=Select;
      Case MenuTypes[i] of
      _INT: Str(LongInt(MenuItems[i]^),S);
      _FLOAT: Str(Real(MenuItems[i]^):0:MenuDigits[i],S);
      End;
      SetColor(Black);
      Rectangle(536,i*28-18,544+Length(S)*8,i*28-4);
      S:=S+'_';
      SetColor(Yellow);
      OutTextXY(540,i*28-14,S);
      Rectangle(536,i*28-18,544+Length(S)*8,i*28-4);
      j:=Length(S); { cursor position }
      Repeat
        NextKey:=ReadKey;
        Case NextKey of
        #0:
          Case ReadKey of
          'K': { left arrow }
            If j>1 then
            Begin
              SetColor(Black);
              OutTextXY(540,i*28-14,S);
              Delete(S,j,1);
              Dec(j);
              Insert('_',S,j);
              SetColor(Yellow);
              OutTextXY(540,i*28-14,S);
            End;
          'M': { right arrow }
            If j<Length(S) then
            Begin
              SetColor(Black);
              OutTextXY(540,i*28-14,S);
              Delete(S,j,1);
              Inc(j);
              Insert('_',S,j);
              SetColor(Yellow);
              OutTextXY(540,i*28-14,S);
            End;
          'S': { Delete }
            If j<Length(S) then
            Begin
              SetColor(Black);
              OutTextXY(540,i*28-14,S);
              Rectangle(536,i*28-18,544+Length(S)*8,i*28-4);
              Delete(s,j+1,1);
              SetColor(Yellow);
              OutTextXY(540,i*28-14,S);
              Rectangle(536,i*28-18,544+Length(S)*8,i*28-4);
            End;
          End;
        '0'..'9','.','-':
          If Length(S)<10 then
          Begin
            SetColor(Black);
            OutTextXY(540,i*28-14,S);
            Rectangle(536,i*28-18,544+Length(S)*8,i*28-4);
            S:=copy(S,1,j-1)+NextKey+'_'+copy(S,j+1,Length(S)-j);
            Inc(j);
            SetColor(Yellow);
            OutTextXY(540,i*28-14,S);
            Rectangle(536,i*28-18,544+Length(S)*8,i*28-4);
          End;
        #8: { BackSpace }
          If j>1 then
          Begin
            SetColor(Black);
            OutTextXY(540,i*28-14,S);
            Rectangle(536,i*28-18,544+Length(S)*8,i*28-4);
            Delete(s,j-1,1);
            Dec(j);
            SetColor(Yellow);
            OutTextXY(540,i*28-14,S);
            Rectangle(536,i*28-18,544+Length(S)*8,i*28-4);
          End;
        #27: Break;
        #13:
          Begin
            Delete(S,j,1);
            Val(S,Temp,ErrCode);
            Insert('_',S,j);
            If ErrCode=0 then
            Begin
              If Temp<MenuMin[i] then Temp:=MenuMin[i];
              If Temp>MenuMax[i] then Temp:=MenuMax[i];
              If MenuTypes[i]=_INT then
                LongInt(MenuItems[i]^):=Trunc(Temp) else
                Real(MenuItems[i]^):=Temp;
              Break;
            End;
          End;
        End;
      Until False;
      SetColor(Black);
      OutTextXY(540,i*28-14,S);
      Rectangle(536,i*28-18,544+Length(S)*8,i*28-4);
      i:=Select;
      Case MenuTypes[i] of
      _INT: Str(LongInt(MenuItems[i]^),S);
      _FLOAT: Str(Real(MenuItems[i]^):0:MenuDigits[i],S);
      End;
      SetColor(White);
      OutTextXY(540,i*28-14,S);
      Rectangle(536,i*28-18,544+Length(S)*8,i*28-4);
      Mouse.Show;
    End;
  Until select>8;

  Mouse.Hide;

  Release(BeforeEvol);

  If select=10 then { EXIT }
  Begin
    ExitProgram:=True;
    Exit;
  End;

  InitAll;
  app.CurPage:=0;
  SetVisualPage(app.CurPage);
  ExitSimulation:=False;
End;

Procedure ClearInputBuffer;
Begin
  Repeat
    Mouse.GetState;
  Until not Mouse.lb and not Mouse.rb; { wait buttons release }
  While KeyPressed do ReadKey; { clear keyboard buffer }
End;

Constructor TEvolApplication.Init;
Begin
  if RegisterBGIdriver(@EGAVGA) < 0 then
  begin
    Writeln('Error registering driver: ',GraphErrorMsg(GraphResult));
    Halt(1);
  end;
  gd:=Detect;
  InitGraph(gd,gm,'');
  ErrCode:=GraphResult;
  if ErrCode<>grOk then
  Begin
    Writeln('Graphics error:', GraphErrorMsg(ErrCode));
    Halt;
  End;
  SetGraphMode(1); { For page flipping }

  Mouse.Reset;
  Mouse.SetXLimit(480,640);
  Mouse.SetYLimit(0,350);
  Mouse.SetPos(560,170);

  evol.InitParams;
  Mark(evol.BeforeEvol);
End;

Procedure TEvolApplication.Run;
Begin
  Repeat
    evol.ShowMenu;
    ClearInputBuffer;
    If evol.ExitProgram then Exit;

    While not evol.ExitSimulation do
    Begin
      Time:=Timer;
      CurPage:=1-CurPage;
      WaitVerticalRetrace;
      SetActivePage(CurPage);
      ClearDevice; SetColor(White); Line(476,0,476,349);
      evol.Move;
      If evol.ShowLeaders then evol.ShowLeadersStats;
      SetVisualPage(CurPage);
      Repeat Until Timer-Time>0;
      Mouse.GetState;
    End;

    ClearInputBuffer;

  Until False;
End;

Destructor TEvolApplication.Done;
Begin
  CloseGraph;
End;

Begin
  App.Init;
  App.Run;
  App.Done;
End.